/* Reduce the font size in data frames - See https://github.com/scverse/cookiecutter-scverse/issues/193 */
div.cell_output table.dataframe {
    font-size: 0.8em;
}

.bd-sidebar .caption-text {
    color: #54c285;
    font-weight: 600;
    text-transform: uppercase;
}
